/*
 * Copyright (c) 2003, The JUNG Authors
 * All rights reserved.
 * 
 * This software is open-source under the BSD license; see either "license.txt"
 * or https://github.com/jrtom/jung/blob/master/LICENSE for a description.
 * 
 */
package edu.uci.ics.jung.samples;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JPanel;
import javax.xml.parsers.ParserConfigurationException;

import com.google.common.base.Functions;
import edu.uci.ics.jung.algorithms.layout.*;
import edu.uci.ics.jung.algorithms.metrics.Metrics;
import edu.uci.ics.jung.algorithms.metrics.Statistics;
import edu.uci.ics.jung.algorithms.scoring.BetweennessCentrality;
import edu.uci.ics.jung.algorithms.shortestpath.DistanceStatistics;
import edu.uci.ics.jung.graph.*;
import edu.uci.ics.jung.graph.event.GraphEvent;
import edu.uci.ics.jung.graph.util.Context;
import edu.uci.ics.jung.visualization.decorators.EdgeShape;
import org.xml.sax.SAXException;

import com.google.common.base.Function;
import com.google.common.base.Supplier;

import edu.uci.ics.jung.io.GraphMLReader;
import edu.uci.ics.jung.visualization.GraphZoomScrollPane;
import edu.uci.ics.jung.visualization.VisualizationViewer;
import edu.uci.ics.jung.visualization.control.AbstractModalGraphMouse;
import edu.uci.ics.jung.visualization.control.CrossoverScalingControl;
import edu.uci.ics.jung.visualization.control.DefaultModalGraphMouse;
import edu.uci.ics.jung.visualization.control.GraphMouseListener;
import edu.uci.ics.jung.visualization.control.ScalingControl;
import edu.uci.ics.jung.visualization.decorators.ToStringLabeller;
import edu.uci.ics.jung.visualization.renderers.GradientVertexRenderer;
import edu.uci.ics.jung.visualization.renderers.Renderer;
import edu.uci.ics.jung.visualization.renderers.BasicVertexLabelRenderer.InsidePositioner;


/**
 * Demonstrates loading (and visualizing) a graph from a GraphML file.
 * 
 * @author Tom Nelson
 * 
 */
public class GraphFromGraphMLDemo {

    /**
     * the visual component and renderer for the graph
     */
    VisualizationViewer<Number, Number> vv;
    
    /**
     * Creates an instance showing a simple graph with controls to demonstrate the zoom features.
     * @param filename the file containing the graph data we're reading
     * @throws ParserConfigurationException if a SAX parser cannot be constructed
     * @throws SAXException if the SAX parser factory cannot be constructed
     * @throws IOException if the file cannot be read
     */
    public GraphFromGraphMLDemo(String filename, GraphType type) throws ParserConfigurationException, SAXException, IOException {
        
    	Supplier<Number> vertexFactory = new Supplier<Number>() {
    		int n = 0;
    		public Number get() { return n++; }
    	};
    	Supplier<Number> edgeFactory = new Supplier<Number>() {
    		int n = 0;
    		public Number get() { return n++; }
    	};


        if(GraphType.DIRECTED.equals(type)) {
            final GraphMLReader<DirectedGraph<Number,Number>, Number, Number> gmlr = new GraphMLReader<DirectedGraph<Number,Number>, Number, Number>(vertexFactory, edgeFactory);
            final DirectedGraph<Number,Number> graph = new DirectedSparseMultigraph<Number,Number>();
            setEnvironment(gmlr, graph, filename);
        } else {
            final GraphMLReader<UndirectedGraph<Number,Number>, Number, Number> gmlr = new GraphMLReader<UndirectedGraph<Number,Number>, Number, Number>(vertexFactory, edgeFactory);
            final UndirectedGraph<Number,Number> graph = new UndirectedSparseGraph<Number, Number>();
            setEnvironment(gmlr, graph, filename);
        }

    }

    private void setEnvironment(GraphMLReader gmlr, Graph<Number,Number> graph, String filename) throws IOException {

        gmlr.load(filename, graph);

        // create a simple graph for the demo
        vv =  new VisualizationViewer<Number,Number>(new SpringLayout2<Number, Number>(graph));
        vv.addGraphMouseListener(new TestGraphMouseListener<Number>());
        vv.getRenderer().setVertexRenderer(
                new GradientVertexRenderer<Number,Number>(
                        Color.white, Color.red,
                        Color.white, Color.blue,
                        vv.getPickedVertexState(),
                        false));

        // add my listeners for ToolTips
        vv.setVertexToolTipTransformer(new ToStringLabeller(){
            @Override
            public String apply(Object o) {
                return super.apply(o);
            }
        });
        vv.getRenderContext().setVertexLabelTransformer(new Function<Number, String>(){
            public String apply(Number o) {
                String s = gmlr.getVertexIDs().get(o).toString();
                return s;
            }
        });
        vv.setEdgeToolTipTransformer(new Function<Number,String>() {
            public String apply(Number edge) {
                return "E"+graph.getEndpoints(edge).toString();
            }});

/*
        vv.getRenderContext().setEdgeLabelTransformer(new Function<Number, String>()  {
            public String apply(Number o) {
                String s = gmlr.getEdgeDescriptions().get(o);
                if(s==null)
                    return "1";
                return s;
            }
        });
*/

        vv.getRenderContext().setEdgeFontTransformer(Functions.constant(new Font("Helvetica", Font.PLAIN, 14)));
        vv.getRenderContext().setVertexFontTransformer(Functions.constant(new Font("Helvetica", Font.PLAIN, 14)));

        // multigraph only!
//        vv.getRenderContext().setEdgeShapeTransformer(EdgeShape.quadCurve(graph));

        //vv.getRenderContext().setVertexLabelTransformer(new ToStringLabeller());
        vv.getRenderer().getVertexLabelRenderer().setPositioner(new InsidePositioner());
        vv.getRenderer().getVertexLabelRenderer().setPosition(Renderer.VertexLabel.Position.AUTO);

        // create a frome to hold the graph
        final JFrame frame = new JFrame();
        Container content = frame.getContentPane();
        final GraphZoomScrollPane panel = new GraphZoomScrollPane(vv);
        content.add(panel);
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        final AbstractModalGraphMouse graphMouse = new DefaultModalGraphMouse<Number,Number>();
        vv.setGraphMouse(graphMouse);
        vv.addKeyListener(graphMouse.getModeKeyListener());

        JMenuBar menubar = new JMenuBar();
        menubar.add(graphMouse.getModeMenu());
        panel.setCorner(menubar);


        vv.addKeyListener(graphMouse.getModeKeyListener());
        vv.setToolTipText("<html><center>Type 'p' for Pick mode<p>Type 't' for Transform mode");

        final ScalingControl scaler = new CrossoverScalingControl();

        JButton plus = new JButton("+");
        plus.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                scaler.scale(vv, 1.1f, vv.getCenter());
            }
        });
        JButton minus = new JButton("-");
        minus.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                scaler.scale(vv, 1/1.1f, vv.getCenter());
            }
        });

        JPanel controls = new JPanel();
        controls.add(plus);
        controls.add(minus);
        content.add(controls, BorderLayout.SOUTH);

        frame.pack();
        frame.setTitle(filename);
        frame.setVisible(true);

        String[] filePath = filename.split("/");
        System.out.println("Graf " + filePath[filePath.length-1]);
        System.out.println("Liczba gałęzi: " + graph.getEdgeCount());
        System.out.println("Liczba wierzchołków: " + graph.getVertexCount());
        System.out.println("Wsp. klasteryzacji (śr, min, max): " + Statistics.mean(Metrics.clusteringCoefficients(graph).values())
                        + ", " + Statistics.min(Metrics.clusteringCoefficients(graph).values())
                        + ", " + Statistics.max(Metrics.clusteringCoefficients(graph).values()));

        ArrayList<Integer> degrees = new ArrayList<>();
        GraphDecorator<Number, Number> g = new GraphDecorator<>(graph);

        for(Number v: g.getVertices()) degrees.add(g.degree(v));
        System.out.println("Sredni stopień: " + Statistics.mean(degrees));

        degrees = new ArrayList<>();
        for(Number v: g.getVertices()) degrees.add(g.inDegree(v));
        System.out.println("Sredni stopień wchodzący: " + Statistics.mean(degrees));

        degrees = new ArrayList<>();
        for(Number v: g.getVertices()) degrees.add(g.outDegree(v));
        System.out.println("Sredni stopień wychodzący: " + Statistics.mean(degrees));

        System.out.println("Srednica: " + DistanceStatistics.diameter(graph));

        int n = graph.getVertexCount();
        int potentialEdgeCount;
        if(graph instanceof DirectedGraph)
            potentialEdgeCount = n*(n-1) + n; //pełny w klasie <1, 0, 1>
        else
            potentialEdgeCount = (n*(n-1))/2 + n; //pełny w klasie <0, 1, 1>

        System.out.println("Gęstość: " + (double)graph.getEdgeCount()/potentialEdgeCount);

        ArrayList<Double> betweenness = new ArrayList<>();
        for(Number v: g.getVertices()) betweenness.add(new BetweennessCentrality(graph).getVertexScore(v));
        System.out.println("Sredni betweenness wierzchołków: " + Statistics.mean(betweenness));

        betweenness = new ArrayList<>();
        for(Number e: g.getEdges()) betweenness.add(new BetweennessCentrality(graph).getEdgeScore(e));
        System.out.println("Sredni betweenness gałęzi: " + Statistics.mean(betweenness));
    }
    
    /**
     * A nested class to demo the GraphMouseListener finding the
     * right vertices after zoom/pan
     */
    static class TestGraphMouseListener<V> implements GraphMouseListener<V> {
        
    		public void graphClicked(V v, MouseEvent me) {
    		    System.err.println("Vertex "+v+" was clicked at ("+me.getX()+","+me.getY()+")");
    		}
    		public void graphPressed(V v, MouseEvent me) {
    		    System.err.println("Vertex "+v+" was pressed at ("+me.getX()+","+me.getY()+")");
    		}
    		public void graphReleased(V v, MouseEvent me) {
    		    System.err.println("Vertex "+v+" was released at ("+me.getX()+","+me.getY()+")");
    		}
    }

    /**
     * @param args if this contains at least one element, the first will be used as the file to read
     * @throws ParserConfigurationException if a SAX parser cannot be constructed
     * @throws SAXException if the SAX parser factory cannot be constructed
     * @throws IOException if the file cannot be read
     */
    public static void main(String[] args) throws ParserConfigurationException, SAXException, IOException 
    {
        loadGraph("datasets/graf_skierowany.graphml", GraphType.DIRECTED);
        loadGraph("datasets/multigraf_skierowany.graphml", GraphType.DIRECTED);
        loadGraph("datasets/siec_skierowana.graphml", GraphType.DIRECTED);

        loadGraph("datasets/repr_uproszczona.graphml", GraphType.DIRECTED);
        loadGraph("datasets/repr_prosta.graphml", GraphType.DIRECTED);
        loadGraph("datasets/repr_pelna.graphml", GraphType.DIRECTED);

        loadGraph("datasets/graf_skierowany.graphml", GraphType.DIRECTED);
        loadGraph("datasets/graf_nieskierowany.graphml", GraphType.UNDIRECTED);

    }

    private static void loadGraph(String resourcePath, GraphType graphType) throws ParserConfigurationException, SAXException, IOException {
        URL fileUrl = SimpleGraphDraw.class.getClassLoader().getResource(resourcePath);
        URI fileUri = null;
        try {
            fileUri =  fileUrl != null ? fileUrl.toURI() : null;
        } catch (URISyntaxException e) {
            e.printStackTrace();
        }
        String filename = fileUri != null ? fileUri.getPath() : null;
        new GraphFromGraphMLDemo(filename, graphType);
    }

    private enum GraphType{
        DIRECTED,
        UNDIRECTED
    }
}
