package pl.edu.wat.wcy.isi.szymonm.phd.musicxml;

import lombok.Getter;
import lombok.Setter;
import pl.edu.wat.wcy.isi.szymonm.phd.enums.notes.NoteConstants;
import pl.edu.wat.wcy.isi.szymonm.phd.enums.notes.NoteNames;
import pl.edu.wat.wcy.isi.szymonm.phd.enums.notes.OctaveNames;
import pl.edu.wat.wcy.isi.szymonm.phd.enums.notes.RhythmValues;

import javax.xml.bind.annotation.*;

@Getter
@Setter
@XmlRootElement(name = "Note")
@XmlAccessorType(XmlAccessType.FIELD)
public class Note {

    @XmlTransient
    private static int NOTE_COUNT = 0;

    @XmlAttribute
    private int pitch;
    @XmlAttribute
    private int dynamic;
    @XmlAttribute
    private double rhythmValue;
    @XmlAttribute
    private double duration;
    @XmlAttribute
    private double pan;

    // transient - field used only for internal calculations
    @XmlTransient
    private double frequency;
    @XmlTransient
    private int noteOrdinal;

    public Note() {
        noteOrdinal = NOTE_COUNT;
        NOTE_COUNT++;
    }

    public void setPitch(int pitch) {
        this.pitch = pitch;
        this.frequency = getFrequencyFromPitch(pitch);
    }

    public void setFrequency(double frequency) {
        this.frequency = frequency;
        this.pitch = getPitchFromFrequency(frequency);
    }

    private int getPitchFromFrequency(double frequency) {
        if (frequency < 0) {
            return -1;
        }
        double keyNumberApprox = 12 * log2(frequency / NoteConstants.A_NOTE_FREQUENCY) + NoteConstants.A_NOTE_PITCH;

        return (int)Math.round(keyNumberApprox);
    }

    private double getFrequencyFromPitch(int pitch) {
        if (pitch < 0) {
            return -1;
        }
        int pitchDistance = pitch - NoteConstants.A_NOTE_PITCH;
        double frequencyMultiplier = Math.pow(NoteConstants.NOTE_FREQ_MULTIPLIER, pitchDistance);

        return NoteConstants.A_NOTE_FREQUENCY * frequencyMultiplier;
    }

    public boolean isPause(){
        return pitch <= 0;
    }

    public RhythmValues getRhythmValuesValue(){
        int maxValue = 32;
        int constantOrdinal = 0;
        if (Math.abs(rhythmValue) < 0.001) {
            rhythmValue = duration;
        }
        if (rhythmValue > 1.5) {
            return RhythmValues.DOTTED_FULL;
        }
        double eps = rhythmValue > 0.2 ? 0.01 : 0.001;

        for (int i = 1; i <= maxValue; i = i*2) {
            double noteLength = 1.0/i;
            double dottedRhythmValue = 1.5 * noteLength;
            final double diffToDotted = Math.abs(rhythmValue - dottedRhythmValue);
            if(diffToDotted < eps){
                printSelectedNoteRhythm(constantOrdinal);
                return RhythmValues.values()[constantOrdinal];
            }
            if(rhythmValue < dottedRhythmValue && rhythmValue > noteLength){
                final int ordinalModifier = determineCloser(rhythmValue, dottedRhythmValue, noteLength);
                final int effectiveConstantOrdinal = constantOrdinal + ordinalModifier;
                printSelectedNoteRhythm(effectiveConstantOrdinal);
                return RhythmValues.values()[effectiveConstantOrdinal];
            }
            constantOrdinal++;

            final double diffToNote = Math.abs(rhythmValue - noteLength);
            if(diffToNote < eps){
                printSelectedNoteRhythm(constantOrdinal);
                return RhythmValues.values()[constantOrdinal];
            }
            double nextDotted = 1.5/(i*2);
            if(rhythmValue > nextDotted && rhythmValue < noteLength){
                final int ordinalModifier = determineCloser(rhythmValue, noteLength, nextDotted);
                final int effectiveConstantOrdinal = constantOrdinal + ordinalModifier;
                printSelectedNoteRhythm(effectiveConstantOrdinal);
                return RhythmValues.values()[effectiveConstantOrdinal];
            }
            constantOrdinal++;
        }
        System.out.println("Nie znaleziono wartości rytmicznej dla nuty nr: " + noteOrdinal
                + " - rhythmValue: " + rhythmValue
                + " duration: " + this.duration + " freq: " + this.frequency + " pitch: " + this.pitch);
        return RhythmValues.UNKNOWN;
    }

    private void printSelectedNoteRhythm(int constantOrdinal) {
        System.out.println("Note " + noteOrdinal + " rhythmValue: " + rhythmValue + " chosen rhythm: " + RhythmValues.values()[constantOrdinal]);
    }

    private int determineCloser(double rhythmValue, double dottedRhythmValue, double noteLength) {
        final double distanceToBigger = Math.abs(rhythmValue - dottedRhythmValue);
        final double distanceToLower = Math.abs(rhythmValue - noteLength);
        return distanceToBigger > distanceToLower ? 1 : 0;
    }

    public String getNoteFullName(){
        return getNoteStandardName() + "_" + getRhythmString();
    }

    public String getNoteStandardName(){
        if(this.pitch < 1)
            return getNoteShortName();
        else if(getNoteOctaveFromFrequency().equals(OctaveNames.Wielka))
            return getNoteShortName(); // C, C#, D, ...
        else if(getNoteOctaveFromFrequency().equals(OctaveNames.Mala))
            return getNoteShortName().toLowerCase(); //c, c#, d, ...
        else
            return getNoteShortName() + getOctaveStringSymbol(); //c₂, ..., c⁶
    }

    public String getNoteShortNameFromPitch(){
        //double keyNumberApprox = 12 * log2((double)this.pitch / 440) + 49;
        if(pitch < 1)
            return "P";

        int noteNameIndex = pitch % 12;
        if(noteNameIndex < 0)
            return "P";

        NoteNames note = NoteNames.values()[noteNameIndex];
        return note.name();
    }

    public String getNoteShortName() {
        return getNoteShortName(3);
    }

    public String getNoteShortNameWithCFirst() {
        return getNoteShortName(7);
    }

    private String getNoteShortName(int noteEnumOffset){
        if (frequency < 1)
            return "P";

        int keyNumber = getNoteIndex() + noteEnumOffset;
        int noteNameIndex = keyNumber % 12;
        if(noteNameIndex < 0)
            return "P";

        NoteNames note = NoteNames.values()[noteNameIndex];
        return note.name();
    }

    public int getNoteIndex(){
        // all rests are mapped to 0
        if (frequency < 1) {
            return 0;
        }

        // formula used for mapping note pitch to piano key number assuming that A1 == 440 Hz is 49th key on the 88-key piano
        double keyNumberApprox = 12 * log2(this.frequency / 440.0) + 49;

        int keyNumber = (int)Math.round(keyNumberApprox);

        // formula below used for maping notes to indices: C to 1, C# to 2, D to 3, ..., H (Bb) to 12
        int noteNameIndex = ((keyNumber + 8) % 12) + 1;
        return Math.max(noteNameIndex, 0);
    }

    public OctaveNames getNoteOctaveFromFrequency() {
        if (this.frequency < 0) {
            return OctaveNames.Mala; // wymusza brak dodawania oktawy do pauz
        }

        if (this.frequency > 15 && this.frequency < 32.7)
            return OctaveNames.Subkontra;
        else if (this.frequency < 65.4)
            return OctaveNames.Kontra;
        else if (this.frequency < 130.8)
            return OctaveNames.Wielka;
        else if (this.frequency < 261.6)
            return OctaveNames.Mala;
        else if (this.frequency <  523.2)
            return OctaveNames.Razkreslna;
        else if (this.frequency < 1046.4)
            return OctaveNames.Dwukreslna;
        else if (this.frequency < 2092.9)
            return OctaveNames.Trzykreslna;
        else if (this.frequency < 4185.9)
            return OctaveNames.Czterokreslna;
        else if (this.frequency < 8371.9)
            return OctaveNames.Pieciokreslna;
        else if (this.frequency < 16745)
            return OctaveNames.Szesciokreslna;
        else throw new IllegalArgumentException("Częstotliwość dźwięku " + frequency + " nie mieści się w rozpatrywanym zakresie!");
    }

    public OctaveNames getNoteOctaveFromPitch() {
        if (this.pitch < 0) {
            return OctaveNames.Mala; // wymusza brak dodawania oktawy do pauz
        }

        if (this.pitch > 15 && this.pitch < 32.7)
            return OctaveNames.Subkontra;
        else if (this.pitch < 65.4)
            return OctaveNames.Kontra;
        else if (this.pitch < 130.8)
            return OctaveNames.Wielka;
        else if (this.pitch < 261.6)
            return OctaveNames.Mala;
        else if (this.pitch <  523.2)
            return OctaveNames.Razkreslna;
        else if (this.pitch < 1046.4)
            return OctaveNames.Dwukreslna;
        else if (this.pitch < 2092.9)
            return OctaveNames.Trzykreslna;
        else if (this.pitch < 4185.9)
            return OctaveNames.Czterokreslna;
        else if (this.pitch < 8371)
            return OctaveNames.Pieciokreslna;
        else if (this.pitch < 16744)
            return OctaveNames.Szesciokreslna;
        else throw new IllegalArgumentException("Częstotliwość dźwięku " + pitch + " nie mieści się w rozpatrywanym zakresie!");
    }

    public OctaveNames getNoteOctave(){

        for(int i=0; i < 9; i++)
            if(pitch < (i+1)*12)
                return OctaveNames.values()[i];

        throw new IllegalArgumentException("Częstotliwość dźwięku " + pitch + " nie mieści się w rozpatrywanym zakresie!");
    }

    public String getOctaveStringSymbol(){
        switch(getNoteOctaveFromFrequency()){
            case Subkontra:
                return "₂";
            case Kontra:
                return "₁";
            case Wielka:
                return "";
            case Mala:
                return "";
            case Razkreslna:
                return "¹";
            case Dwukreslna:
                return "²";
            case Trzykreslna:
                return "³";
            case Czterokreslna:
                return "⁴";
            case Pieciokreslna:
                return "⁵";
            case Szesciokreslna:
                return "⁶";
            default: throw new IllegalArgumentException("Częstotliwość dźwięku " + pitch + " nie należy do żadnej z oktaw!");
        }
    }

    public OctaveNames getOctaveFromStringSymbol(String noteName, String symbol){
        switch(symbol){
            case "₂":
                return OctaveNames.Subkontra;
            case "₁":
                return OctaveNames.Kontra;
            case "": {
                if (noteName.equals(noteName.toUpperCase()))
                    return OctaveNames.Wielka;
                else
                    return OctaveNames.Mala;
            }
            case "¹":
                return OctaveNames.Razkreslna;
            case "²":
                return OctaveNames.Dwukreslna;
            case "³":
                return OctaveNames.Trzykreslna;
            case "⁴":
                return OctaveNames.Czterokreslna;
            case "⁵":
                return OctaveNames.Pieciokreslna;
            case "⁶":
                return OctaveNames.Szesciokreslna;
            default:
                throw new IllegalArgumentException("Częstotliwość dźwięku " + pitch +
                    " (noteName = " + noteName + ", symbol = " + symbol +
                    ") nie należy do żadnej z oktaw!");
        }
    }

    private String getRhythmString() {
        switch (getRhythmValuesValue()) {
            case DOTTED_FULL:
                return "3/2";
            case FULL:
                return "1/1";
            case DOTTED_SEMI:
                return "3/4";
            case SEMI:
                return "1/2";
            case DOTTED_QUARTER:
                return "3/8";
            case QUARTER:
                return "1/4";
            case DOTTED_EIGHTH:
                return "3/16";
            case EIGHTH:
                return "1/8";
            case DOTTED_SIXTEENTH:
                return "3/32";
            case SIXTEENTH:
                return "1/16";
            case DOTTED_THIRTY_SECONDTH:
                return "3/64";
            case THIRTY_SECOND:
                return "1/32";
            default:
                return "?";
        }
    }

    public String getRhythmString(RhythmValues rhythmValue) {
        switch (rhythmValue) {
            case DOTTED_FULL:
                return "3/2";
            case FULL:
                return "1/1";
            case DOTTED_SEMI:
                return "3/4";
            case SEMI:
                return "1/2";
            case DOTTED_QUARTER:
                return "3/8";
            case QUARTER:
                return "1/4";
            case DOTTED_EIGHTH:
                return "3/16";
            case EIGHTH:
                return "1/8";
            case DOTTED_SIXTEENTH:
                return "3/32";
            case SIXTEENTH:
                return "1/16";
            case DOTTED_THIRTY_SECONDTH:
                return "3/64";
            case THIRTY_SECOND:
                return "1/32";
            default:
                System.out.println("Nie znaleziono wartości dla nuty " + noteOrdinal + " rhythmValue: " + this.rhythmValue + " przekazana wartość: " + rhythmValue);
                return "?";
        }
    }

    public double getDurationFromRhythmString(String rhythmString) {
        switch (rhythmString) {
            case "3/2":
                return (double) 3 / 2;
            case "1/1":
                return 1;
            case "3/4":
                return (double) 3 / 4;
            case "1/2":
                return (double) 1 / 2;
            case "3/8":
                return (double) 3 / 8;
            case "1/4":
                return (double) 1 / 4;
            case "3/16":
                return (double) 3 / 16;
            case "1/8":
                return (double) 1 / 8;
            case "3/32":
                return (double) 3 / 32;
            case "1/16":
                return (double) 1 / 16;
            case "3/64":
                return (double) 3 / 64;
            case "1/32":
                return (double) 1 / 32;
            default:
                System.out.println("Nie znaleziono wartości dla nuty " + noteOrdinal + " rhythmValue: " + rhythmValue + " przekazana wartość: " + rhythmString);
                return 1;
        }
    }

    private double log2(double value){
        return Math.log(value)/Math.log(2);
    }
}
